"""Accurate now() based on cached offset between NTP server
and local system time.

Make sure not to change system time once the offset is cached.
"""

import logging
from datetime import datetime, timezone

import ntplib

logger = logging.getLogger(__name__)

DEFAULT_NTP_SERVER = "pool.ntp.org"
cached_offset_between_ntp_server_and_local_system = None


def now_utc(ntp_server=DEFAULT_NTP_SERVER):
    global cached_offset_between_ntp_server_and_local_system

    if cached_offset_between_ntp_server_and_local_system is not None:
        return (
            datetime.now(timezone.utc)
            + cached_offset_between_ntp_server_and_local_system
        )

    try:
        resp = ntplib.NTPClient().request(ntp_server)
        adjusted_timestamp = resp.tx_time + resp.delay * 0.5
        ntp_server_time = datetime.fromtimestamp(adjusted_timestamp, timezone.utc)
        # update cache
        cached_offset_between_ntp_server_and_local_system = (
            ntp_server_time - datetime.now(timezone.utc)
        )
        logger.debug(
            "Successfully retrieved time from NTP server {srv}. time:{time}, offset:{offset}".format(
                srv=ntp_server,
                time=ntp_server_time,
                offset=cached_offset_between_ntp_server_and_local_system.total_seconds(),
            )
        )
        return ntp_server_time

    except Exception as e:
        logger.debug(
            "Failed to retrieve time from NTP server {srv}. error {err}".format(
                srv=ntp_server, err=str(e)
            )
        )

    return datetime.now(timezone.utc)


def reset_cached_offset():
    global cached_offset_between_ntp_server_and_local_system
    cached_offset_between_ntp_server_and_local_system = None


def get_cached_offset():
    global cached_offset_between_ntp_server_and_local_system
    return cached_offset_between_ntp_server_and_local_system



from binascii import hexlify

from autouri.metadata import base64_to_hex, get_seconds_from_epoch, parse_md5_str


def test_get_seconds_from_epoch():
    assert get_seconds_from_epoch("Sat, 07 Mar 2020 21:03:07 GMT") == 1583614987.0
    assert get_seconds_from_epoch("Sat Mar  7 15:05:07 PST 2020") == 1583622307.0
    assert get_seconds_from_epoch("2017-07-01T14:59:55.711Z") == 1498921195.711
    assert get_seconds_from_epoch("2017-03-10 14:30:12,655+0000") == 1489156212.655


def test_base64_to_hex():
    assert base64_to_hex("aGVsbG8gd29ybGQ=") == hexlify("hello world".encode()).decode()
    assert (
        base64_to_hex("YXNkZnNkZnNhZGYjQEtKV0xLSkFTRDw+S0ZKQCNBU0RDQ0FTREBAIyRSIQ==")
        == hexlify("asdfsdfsadf#@KJWLKJASD<>KFJ@#ASDCCASD@@#$R!".encode()).decode()
    )
    assert (
        base64_to_hex("NWEyNjUzYTY2Zjk1OGY5ZmNiODYwMGUyMDI4MTFiMjc=")
        == hexlify("5a2653a66f958f9fcb8600e202811b27".encode()).decode()
    )


def test_parse_md5_str():
    assert (
        parse_md5_str("WiZTpm+Vj5/LhgDiAoEbJw==") == "5a2653a66f958f9fcb8600e202811b27"
    )
    assert (
        parse_md5_str("5a2653a66f958f9fcb8600e202811b27")
        == "5a2653a66f958f9fcb8600e202811b27"
    )



import os
import time
from typing import Any, Tuple

import pytest
from requests.exceptions import HTTPError

from autouri.autouri import AutoURI, URIBase
from autouri.httpurl import HTTPURL, ReadOnlyStorageError
from autouri.s3uri import S3URI

from .files import (
    common_paths,
    make_files_in_dir,
    recurse_raise_if_uri_not_exist,
    v6_txt_contents,
)


@pytest.mark.parametrize("path", common_paths())
def test_s3uri_uri(path) -> Any:
    assert S3URI(path).uri == path


@pytest.mark.parametrize("path", common_paths())
def test_s3uri_uri_wo_ext(path) -> str:
    assert S3URI(path).uri_wo_ext == os.path.splitext(path)[0]


@pytest.mark.parametrize("path", common_paths())
def test_s3uri_uri_wo_scheme(path) -> str:
    assert S3URI(path).uri_wo_scheme == path.replace("s3://", "", 1)


@pytest.mark.parametrize("path", common_paths())
def test_s3uri_is_valid(path) -> bool:
    """Also tests AutoURI auto-conversion since it's based on is_valid property"""
    expected = path.startswith("s3://")
    assert S3URI(path).is_valid == expected
    assert not expected or type(AutoURI(path)) == S3URI


@pytest.mark.parametrize("path", common_paths())
def test_s3uri_dirname(path) -> str:
    assert S3URI(path).dirname == os.path.dirname(path)


@pytest.mark.parametrize("path", common_paths())
def test_s3uri_dirname_wo_scheme(path) -> str:
    assert S3URI(path).dirname_wo_scheme == os.path.dirname(path).replace(
        "s3://", "", 1
    )


@pytest.mark.parametrize("path", common_paths())
def test_s3uri_loc_dirname(path) -> str:
    assert S3URI(path).loc_dirname == os.path.dirname(path).replace(
        "s3://", "", 1
    ).lstrip("/")


@pytest.mark.parametrize("path", common_paths())
def test_s3uri_basename(path) -> str:
    assert S3URI(path).basename == os.path.basename(path)


@pytest.mark.parametrize("path", common_paths())
def test_s3uri_basename_wo_ext(path) -> str:
    assert S3URI(path).basename_wo_ext == os.path.splitext(os.path.basename(path))[0]


@pytest.mark.parametrize("path", common_paths())
def test_s3uri_ext(path) -> str:
    assert S3URI(path).ext == os.path.splitext(path)[1]


def test_s3uri_exists(s3_v6_txt):
    assert S3URI(s3_v6_txt).exists
    assert not S3URI(s3_v6_txt + ".should-not-be-here").exists
    assert not S3URI("s3://hey/this/should/not/be/here.txt").exists


def test_s3uri_mtime(s3_v6_txt):
    u = S3URI(s3_v6_txt + ".tmp")
    u.write("temp file for testing")
    now = time.time()
    assert now - 10 < u.mtime < now + 10
    u.rm()
    assert not u.exists


def test_s3uri_size(s3_v6_txt, v6_txt_size):
    assert S3URI(s3_v6_txt).size == v6_txt_size


def test_s3uri_md5(s3_v6_txt, v6_txt_md5_hash):
    assert S3URI(s3_v6_txt).md5 == v6_txt_md5_hash


def test_s3uri_md5_from_file(s3_v6_txt, v6_txt_md5_hash):
    u_md5 = S3URI(s3_v6_txt + URIBase.MD5_FILE_EXT)
    if u_md5.exists:
        u_md5.rm()
    assert not u_md5.exists
    u = S3URI(s3_v6_txt)
    assert u.md5_from_file is None

    u.get_metadata(make_md5_file=True)
    # S3URI should not make md5 file even with make_md5_file=True
    assert not u_md5.exists
    assert u.md5_from_file is None

    # nevertheless AutoURI should be able to read from s3://*.md5
    # make a temporary .md5 file
    u_md5.write(v6_txt_md5_hash)
    assert u.md5_from_file == v6_txt_md5_hash
    u_md5.rm()


def test_s3uri_md5_file_uri(s3_v6_txt):
    assert (
        S3URI(s3_v6_txt + URIBase.MD5_FILE_EXT).uri == s3_v6_txt + URIBase.MD5_FILE_EXT
    )


def test_s3uri_cp_url(s3_v6_txt, url_test_path) -> "AutoURI":
    """Test copying local_v6_txt to the following destination storages:
    url_test_path: s3 -> url
        This will fail as intended since URL is read-only.
    """
    u = S3URI(s3_v6_txt)
    basename = os.path.basename(s3_v6_txt)

    for test_path in (url_test_path,):
        u_dest = AutoURI(os.path.join(test_path, "test_s3uri_cp", basename))

        with pytest.raises(ReadOnlyStorageError):
            _, ret = u.cp(u_dest, return_flag=True)


def test_s3uri_cp(s3_v6_txt, local_test_path, s3_test_path, gcs_test_path) -> "AutoURI":
    """Test copying local_v6_txt to the following destination storages:
        local_test_path: s3 -> local
        s3_test_path: s3 -> s3
        gcs_test_path: s3 -> gcs

    Parameters to be tested:
        no_lock:
            Copy with no locking mechanism. There is no way to test this thoroughly here.
            This will be tested with multiple threads later in test_rece_cond.py.
        no_checksum:
            Don't check md5-hash/size/mtime to skip copying (even if file already exists on destination).
        make_md5_file:
            Make md5 file on destination only when it's REQUIRED.
            It's required only if we need to compare md5 hash of source and target.
    """
    u = S3URI(s3_v6_txt)
    basename = os.path.basename(s3_v6_txt)

    for test_path in (local_test_path, s3_test_path, gcs_test_path):
        u_dest = AutoURI(os.path.join(test_path, "test_s3uri_cp", basename))
        if u_dest.exists:
            u_dest.rm()

        assert not u_dest.exists
        _, ret = u.cp(u_dest, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 0
        u_dest.rm()

        assert not u_dest.exists
        # cp without lock will be tested throughly in test_race_cond.py
        _, ret = u.cp(u_dest, no_lock=True, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 0
        u_dest.rm()

        # trivial: copy without checksum when target doesn't exists
        assert not u_dest.exists
        _, ret = u.cp(u_dest, no_checksum=True, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 0

        # copy without checksum when target exists
        m_dest = u_dest.get_metadata()
        assert m_dest.exists
        time.sleep(1)
        _, ret = u.cp(u_dest, no_checksum=True, return_flag=True)
        # compare new mtime vs old mtime
        # new time should be larger if it's overwritten as intended
        assert u_dest.mtime > m_dest.mtime and u.read() == u_dest.read() and ret == 0

        # copy with checksum when target exists
        m_dest = u_dest.get_metadata()
        assert m_dest.exists
        _, ret = u.cp(u_dest, return_flag=True)
        # compare new mtime vs old mtime
        # new time should be the same as old time
        assert u_dest.mtime == m_dest.mtime and u.read() == u_dest.read() and ret == 1

        # make_md5_file works only when it's required
        # i.e. when we need to compare md5 hash of src vs target
        # so target must exist prior to test it
        assert u_dest.exists
        # delete md5 file if exists
        u_dest_md5_file = AutoURI(u_dest.uri + URIBase.MD5_FILE_EXT)
        if u_dest_md5_file.exists:
            u_dest_md5_file.rm()
        _, ret = u.cp(u_dest, make_md5_file=True, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 1
        u_dest.rm()


def test_s3uri_write(s3_test_path):
    u = S3URI(s3_test_path + "/test_s3uri_write.tmp")

    assert not u.exists
    u.write("test")
    assert u.exists and u.read() == "test"
    u.rm()

    # this will be tested more with multiple threads in test_race_cond.py
    assert not u.exists
    u.write("test2", no_lock=True)
    assert u.exists and u.read() == "test2"
    u.rm()
    assert not u.exists


def test_s3uri_rm(s3_test_path):
    u = S3URI(s3_test_path + "/test_s3uri_rm.tmp")

    assert not u.exists
    u.write("")
    assert u.exists
    u.rm()
    assert not u.exists

    # this will be tested more with multiple threads in test_race_cond.py
    assert not u.exists
    u.write("", no_lock=True)
    assert u.exists
    u.rm()
    assert not u.exists


def test_s3uri_get_metadata(s3_v6_txt, v6_txt_size, v6_txt_md5_hash):
    u = S3URI(s3_v6_txt)

    m1 = u.get_metadata()
    assert m1.md5 == v6_txt_md5_hash
    assert m1.size == v6_txt_size

    m2 = u.get_metadata(skip_md5=True)
    assert m2.md5 is None
    assert m2.size == v6_txt_size

    u_md5 = S3URI(s3_v6_txt + ".md5")
    if u_md5.exists:
        u_md5.rm()
    m3 = u.get_metadata(make_md5_file=True)
    assert m3.md5 == v6_txt_md5_hash
    assert m3.size == v6_txt_size
    # S3URI should not make md5 file even with make_md5_file=True
    assert not u_md5.exists


def test_s3uri_read(s3_v6_txt):
    u = S3URI(s3_v6_txt)
    assert u.read() == v6_txt_contents()
    assert u.read(byte=True) == v6_txt_contents().encode()


def test_s3uri_find_all_files(s3_test_path):
    """Make a directory structure with empty files.

    Check if find_all_files() returns correct file (not sub-directory) paths.
    """
    prefix = os.path.join(s3_test_path, "test_s3uri_find_all_files")
    all_files = make_files_in_dir(prefix, make_local_empty_dir_d_a=False)

    all_files_found = S3URI(prefix).find_all_files()
    assert sorted(all_files_found) == sorted(all_files)
    for file in all_files:
        assert S3URI(file).exists


def test_s3uri_rmdir(s3_test_path):
    """Make a directory structure with empty files.

    Check if rmdir() deletes all empty files on given $prefix.
    """
    prefix = os.path.join(s3_test_path, "test_s3uri_rmdir")
    all_files = make_files_in_dir(prefix, make_local_empty_dir_d_a=False)

    # test rmdir(dry_run=True)
    S3URI(prefix).rmdir(dry_run=True)
    for file in all_files:
        assert S3URI(file).exists

    # test rmdir(dry_run=False)
    S3URI(prefix).rmdir(dry_run=False)
    for file in all_files:
        assert not S3URI(file).exists


# original methods in S3URI
def test_s3uri_get_bucket_path():
    assert S3URI("s3://a/b/c/d/e.txt").get_bucket_path() == ("a", "b/c/d/e.txt")
    assert S3URI("s3://asdflskfjljkfc-asdf/ddfjlfd/d.log").get_bucket_path() == (
        "asdflskfjljkfc-asdf",
        "ddfjlfd/d.log",
    )
    assert S3URI("s3://ok-test-bucket/hello.txt").get_bucket_path() == (
        "ok-test-bucket",
        "hello.txt",
    )


def test_s3uri_get_presigned_url(s3_v6_txt):
    u = S3URI(s3_v6_txt)
    # 2 seconds duration
    url = u.get_presigned_url(duration=2)

    u_url = HTTPURL(url)
    assert u_url.is_valid and u_url.read() == v6_txt_contents()
    time.sleep(5)
    # should expire in 2 seconds
    with pytest.raises(HTTPError):
        # forbidden since it's already expired
        u_url.read()


def test_s3uri_get_public_url(s3_public_url_test_v6_file):
    url = S3URI(s3_public_url_test_v6_file).get_public_url()
    u_url = HTTPURL(url)
    assert u_url.is_valid
    assert u_url.read() == v6_txt_contents()


# classmethods
def test_s3uri_get_path_sep() -> str:
    assert S3URI.get_path_sep() == os.path.sep


def test_s3uri_get_schemes() -> Tuple[str, ...]:
    assert S3URI.get_schemes() == ("s3://",)


def test_s3uri_get_loc_suffix() -> str:
    assert S3URI.get_loc_suffix() == ".s3"


def test_s3uri_get_loc_prefix() -> str:
    test_loc_prefix = "test_s3uri_get_loc_prefix"
    S3URI.init_s3uri(loc_prefix=test_loc_prefix)
    assert S3URI.get_loc_prefix() == test_loc_prefix
    S3URI.init_s3uri(loc_prefix="")
    assert S3URI.get_loc_prefix() == ""


def test_s3uri_localize(
    s3_test_path,
    local_j1_json,
    local_v41_json,
    local_v421_tsv,
    local_v5_csv,
    local_v6_txt,
    s3_j1_json,
    s3_v41_json,
    s3_v421_tsv,
    s3_v5_csv,
    s3_v6_txt,
    gcs_j1_json,
    gcs_v41_json,
    gcs_v421_tsv,
    gcs_v5_csv,
    gcs_v6_txt,
    url_j1_json,
    url_v41_json,
    url_v421_tsv,
    url_v5_csv,
    url_v6_txt,
) -> Tuple[str, bool]:
    """Recursive localization is supported for the following file extensions:
        .json:
            Files defined only in values (not keys) can be recursively localized.
        .tsv/.csv:
            Files defined in all values can be recursively localized.

    This function will test localizing j1.json file on each remote storage.
    This JSON file has file paths including .tsv and .csv, which also include
    other files in its contents.
    Therefore, when the recursive flag is on, all files in these JSON, TSV, CSV
    files should be localized recursively with correct file names
    (controlled by cls.loc_prefix and cls.loc_suffix).

    Filenaming for (recursive) localization:
        cls.loc_prefix + remote_file_path_without_scheme + cls.loc_suffix (for recursvely only)

    For example,
    s3://test-bucket/j1.json has some file paths on s3://.

    With recursive localization, all these files must be localized on /tmp/user/loc_prefix/ with
    a correct directory structure (keeping original structure on source: i.e. bucket name, path)
    and the name of the JSON file should be j1.local.json since contents of this file should be
    modified to point to localized files in it. This is recursively done for all files in it too.

    Without recursive localization, autouri doesn't look inside that JSON file and just localize
    the file itself alone on /tmp/user/loc_prefix/ while keeping the same filename j1.local.json.

    Test localizing on a GCS storage from the following remote storages:
        local_test_path: local -> s3
        s3_test_path: s3 -> s3
        gcs_test_path: gcs -> s3
        url_test_path: url -> s3

    Parameters to be tested:
        make_md5_file:
            Make md5 file on destination only when it's REQUIRED.
            It's required only if we need to compare md5 hash of source and target.
            This is already tested cp and it's actually needed for local storage.
            Cloud URIs will provide md5 hash info in their metadata so md5 file
            is not required and hence will not be created even with this flag on.
        recursive:
            j1.json
    """
    loc_prefix = os.path.join(s3_test_path, "test_s3uri_localize")

    for j1_json in (s3_j1_json,):
        # localization from same storage
        u_j1_json = AutoURI(j1_json)
        loc_prefix_ = loc_prefix + u_j1_json.__class__.get_loc_suffix()
        basename = u_j1_json.basename

        # for localization both with or without recursive
        # nothing should be localized actually
        # since they are already on same storage
        # so loc_prefix directory itself shouldn't be created
        loc_uri, localized = S3URI.localize(
            u_j1_json, recursive=False, return_flag=True, loc_prefix=loc_prefix_
        )
        assert loc_uri == u_j1_json.uri and not localized
        assert not AutoURI(os.path.join(loc_prefix_, basename)).exists

        loc_uri, localized = S3URI.localize(
            u_j1_json, recursive=True, return_flag=True, loc_prefix=loc_prefix_
        )
        assert loc_uri == u_j1_json.uri and not localized
        assert not AutoURI(os.path.join(loc_prefix_, basename)).exists
        # check if all URIs defeind in localized JSON file exist
        recurse_raise_if_uri_not_exist(loc_uri)

    # localization from remote storages
    for j1_json in (local_j1_json, gcs_j1_json, url_j1_json):
        u_j1_json = AutoURI(j1_json)
        loc_prefix_ = loc_prefix + u_j1_json.__class__.get_loc_suffix()
        basename = u_j1_json.basename

        loc_uri, localized = S3URI.localize(
            u_j1_json, recursive=False, return_flag=True, loc_prefix=loc_prefix_
        )
        expected = os.path.join(loc_prefix_, u_j1_json.loc_dirname, u_j1_json.basename)
        assert loc_uri == expected
        assert localized and AutoURI(expected).exists

        loc_uri, localized = S3URI.localize(
            u_j1_json, recursive=True, return_flag=True, loc_prefix=loc_prefix_
        )
        expected = os.path.join(
            loc_prefix_,
            u_j1_json.loc_dirname,
            u_j1_json.basename_wo_ext + S3URI.get_loc_suffix() + u_j1_json.ext,
        )
        assert loc_uri == expected
        assert localized and AutoURI(expected).exists
        # check if all URIs defeind in localized JSON file exist
        recurse_raise_if_uri_not_exist(loc_uri)



import logging
import multiprocessing
import os
import uuid
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from filelock import BaseFileLock

from .loc_aux import recurse_csv, recurse_json, recurse_tsv
from .metadata import URIMetadata

logger = logging.getLogger(__name__)


def autouri_rm(uri, thread_id, no_lock=False):
    """Wrapper for AutoURI(uri).rm().
    This function is used for multiprocessing.map() which requires a picklable function
    outside the scope of the class.
    """
    AutoURI(uri, thread_id=thread_id).rm(no_lock=no_lock)


class AutoURIRecursionError(RuntimeError):
    pass


class URIBase(ABC):
    """A base actract class for all URI classes.
    This class is for file only except for the following two methods.
        - list(): Recursively list all files (full path URIs) on a directory.
        - rmdir(): Recursively remove all files on a directory.

    This class can localize (recursively) URI on different URI class.
    (loc. stands for localization.)
    A default localization strategy keeps the directory structure and basename of an original URI.

    Class constants:
        MD5_FILE_EXT:
            File extention for md5 (.md5).
        LOC_RECURSE_EXT_AND_FNC:
            Dict of (file extension, function) to recurse localization.
                e.g. {'.json': recurse_dict, '.tsv': recurse_tsv}
        LOC_RECURSION_DEPTH_LIMIT:
            Limit depth of recursive localization
            to prevent/detect direct/indirect self referencing.
        LOCK_FILE_EXT:
            Lock file's extention (.lock).
        LOCK_TIMEOUT:
            Lock file timeout (-1 for no timeout).
        LOCK_POLL_INTERVAL:
            Lock file polling interval in seconds.

    Class constants for subclasses only
        LOC_PREFIX:
            Cache path prefix for localization on this class' storage.
            This should be '' for this base class
            but must be specified for subclasses.

    Protected class constants:
        _PATH_SEP:
            Separator string for directory.
        _SCHEMES:
            Tuple of scheme strings
        _LOC_SUFFIX:
            Suffix for a localized file
            if file is modified during recursive localization
    """

    MD5_FILE_EXT: str = ".md5"
    LOC_RECURSE_EXT_AND_FNC: Dict[str, Callable] = {
        ".json": recurse_json,
        ".tsv": recurse_tsv,
        ".csv": recurse_csv,
    }

    LOCK_FILE_EXT: str = ".lock"
    LOCK_TIMEOUT: int = 900
    LOCK_POLL_INTERVAL: float = 10.0

    LOC_PREFIX: str = ""
    LOC_RECURSION_DEPTH_LIMIT: int = 10

    DEFAULT_NUM_THREADS = 6

    _PATH_SEP: str = "/"
    _SCHEMES: Tuple[str, ...] = tuple()
    _LOC_SUFFIX: str = ""

    def __init__(self, uri, thread_id=-1):
        if isinstance(uri, URIBase):
            self._uri = uri.uri
        else:
            self._uri = uri
        self._thread_id = thread_id
        self._uuid = str(uuid.uuid4())

    def __repr__(self):
        return self._uri

    def __str__(self):
        return str(self._uri)

    @property
    def thread_id(self):
        """Use for some URI types which are not thread-safe.
        e.g. GCSURI.
        """
        return self._thread_id

    @thread_id.setter
    def thread_id(self, i):
        self._thread_id = i

    @property
    def uuid(self):
        return self._uuid

    @property
    def short_uuid(self):
        return self._uuid[:8]

    @property
    def uri(self) -> Any:
        """Can store any type of variable."""
        return self._uri

    @property
    def uri_wo_ext(self) -> str:
        return os.path.splitext(str(self._uri))[0]

    @property
    def uri_wo_scheme(self) -> str:
        for s in self.__class__.get_schemes():
            if s and str(self._uri).startswith(s):
                return str(self._uri).replace(s, "", 1)
        return str(self._uri)

    @property
    def is_valid(self) -> bool:
        for s in self.__class__.get_schemes():
            if s and str(self._uri).startswith(s):
                return True
        return False

    @property
    def dirname(self) -> str:
        """Dirname with a scheme (gs://, s3://, http://, /, ...)."""
        return os.path.dirname(str(self._uri))

    @property
    def dirname_wo_scheme(self) -> str:
        """Dirname without a scheme (gs://, s3://, http://, /, ...)."""
        return os.path.dirname(self.uri_wo_scheme)

    @property
    def loc_dirname(self) -> str:
        """Dirname to be appended to target cls' LOC_PREFIX after localization.
        Such dirname is stripped with a path separator (e.g. /).

        e.g. localization of src_uri on target cls
            = cls.LOC_PREFIX + src_uri.loc_dirname + src_uri.basename
        """
        return self.dirname_wo_scheme.strip(self.__class__.get_path_sep())

    @property
    def basename(self) -> str:
        """Basename."""
        return os.path.basename(str(self._uri))

    @property
    def basename_wo_ext(self) -> str:
        """Basename without extension."""
        return os.path.splitext(self.basename)[0]

    @property
    def ext(self) -> str:
        """File extension."""
        return os.path.splitext(self.basename)[1]

    @property
    def exists(self) -> bool:
        return self.get_metadata(skip_md5=True).exists

    @property
    def mtime(self) -> float:
        """Seconds since the epoch."""
        return self.get_metadata(skip_md5=True).mtime

    @property
    def size(self) -> int:
        """Size in bytes."""
        return self.get_metadata(skip_md5=True).size

    @property
    def md5(self) -> str:
        """Md5 hash hexadecimal digest string."""
        return self.get_metadata().md5

    @property
    def md5_from_file(self) -> str:
        """Get md5 from a md5 file (.md5) if it exists.
        Check md5 file is newer than the file that it is associated with
        """
        u_md5 = self.md5_file_uri
        if u_md5.is_valid:
            try:
                m_md5 = u_md5.get_metadata(skip_md5=True)
                if m_md5.exists:
                    self_mtime = self.mtime
                    logger.debug(
                        "md5 file exists. mt={mt}, md5_mt={md5_mt}, uri={u}".format(
                            mt=self_mtime, md5_mt=m_md5.mtime, u=self._uri
                        )
                    )
                    if (
                        m_md5.mtime is not None
                        and self_mtime is not None
                        and m_md5.mtime >= self_mtime
                    ):
                        return u_md5.read()
            except Exception:
                pass

        logger.debug(
            "Failed to get md5 hash from md5 file. uri={u}".format(u=self._uri)
        )
        return None

    @property
    def md5_file_uri(self) -> "AutoURI":
        """Get md5 file URI. Not guaranteed to exist"""
        return AutoURI(str(self._uri) + AutoURI.MD5_FILE_EXT)

    def cp(
        self,
        dest_uri: Union[str, "AutoURI"],
        no_lock=False,
        no_checksum=False,
        make_md5_file=False,
        return_flag=False,
    ) -> "AutoURI":
        """Makes a copy on destination. It is protected by a locking mechanism.
        Check md5 hash, file name/size and last modified date if possible to prevent
        unnecessary re-uploading.

        Args:
            dest_uri:
                Target URI.
                If it's an explicit directory with slash (or os.sep) then
                make it a file URI by suffixing self.basename.
            no_lock:
                Do not use a locking mechanism
            no_checksum:
                Do not check md5 hash
            make_md5_file:
                Make an md5 file on destination if metadata doesn't have md5
                assuming that you have write permission on target's directory
                This flag will work only if no_checksum is activated.
                This flag will work only when it's required.
                If source/target storage's get_metadata() can already get md5
                hash then md5 file will not be created.

        Returns:
            Tuple of (s, rc)
                s:
                    URI string of copy on destination
                rc:
                    0:
                        made a copy
                    1:
                        skipped due to same md5 hash
                    2:
                        md5 not found but matched file size and mtime is not newer
        """
        d = AutoURI(dest_uri)
        sep = d.__class__.get_path_sep()
        if d._uri.endswith(sep):
            d = AutoURI(sep.join([d._uri.rstrip(sep), self.basename]))

        logger.info(
            "cp: ({uuid}) started. src={src}, dest={dest}".format(
                uuid=self.short_uuid, src=self._uri, dest=d.uri
            )
        )

        with d.get_lock(no_lock=no_lock):
            if not no_checksum:
                # checksum (by md5, size, mdate)
                m_dest = d.get_metadata(make_md5_file=make_md5_file)
                logger.debug(
                    "cp: ({uuid}) dest metadata={m}, dest={dest}".format(
                        uuid=self.short_uuid, m=m_dest, dest=d.uri
                    )
                )

                if m_dest.exists:
                    m_src = self.get_metadata()
                    logger.debug(
                        "cp: ({uuid}) src metadata={m}, src={src}".format(
                            uuid=self.short_uuid, m=m_src, src=self._uri
                        )
                    )

                    md5_matched = (
                        m_src.md5 is not None
                        and m_dest.md5 is not None
                        and m_src.md5 == m_dest.md5
                    )
                    if md5_matched:
                        logger.info(
                            "cp: ({uuid}) skipped due to md5_match. "
                            "md5={md5}".format(uuid=self.short_uuid, md5=m_src.md5)
                        )
                        return (d._uri, 1) if return_flag else d._uri

                    name_matched = self.basename == d.basename
                    size_matched = (
                        m_src.size is not None
                        and m_dest.size is not None
                        and m_src.size == m_dest.size
                    )
                    src_is_not_newer = (
                        m_src.mtime is not None
                        and m_dest.mtime is not None
                        and m_src.mtime <= m_dest.mtime
                    )
                    if name_matched and size_matched and src_is_not_newer:
                        logger.info(
                            "cp: ({uuid}) skipped due to name_size_match. "
                            "size={sz}, mt={mt}".format(
                                uuid=self.short_uuid, sz=m_src.size, mt=m_src.mtime
                            )
                        )
                        return (d._uri, 2) if return_flag else d._uri

            if not self._cp(dest_uri=d):
                if not d._cp_from(src_uri=self):
                    raise Exception("cp: ({uuid}) failed.".format(uuid=self.short_uuid))

        logger.info("cp: ({uuid}) done.".format(uuid=self.short_uuid))
        return (d._uri, 0) if return_flag else d._uri

    def write(self, s, no_lock=False):
        """Write string/bytes to file. It is protected by a locking mechanism."""
        with self.get_lock(no_lock=no_lock):
            self._write(s)
        return

    def rm(self, no_lock=False, silent=False):
        """Remove a URI from its storage. It is protected by by a locking mechanism."""
        with self.get_lock(no_lock=no_lock):
            self._rm()
            if not silent:
                logger.info(
                    "rm: ({uuid}) {uri}".format(uuid=self.short_uuid, uri=self._uri)
                )
        return

    def rmdir(self, dry_run=False, num_threads=DEFAULT_NUM_THREADS, no_lock=False):
        """Recursively delete all files on a directory.
        Use it at your own risk. This method is multi-threaded.

        Args:
            dry_run:
                Dry-run.
            num_threads:
                Number of threads for deletion.
        """
        files = self.find_all_files()
        if dry_run:
            for uri in files:
                logger.info(
                    "rm (dry-run): ({uuid}) {uri}".format(uuid=self.short_uuid, uri=uri)
                )
            return
        num_files = len(files)
        thread_ids = [i % num_threads for i in range(num_files)]
        no_locks = [no_lock] * num_files

        args = list(zip(files, thread_ids, no_locks))
        with multiprocessing.Pool(num_threads) as p:
            p.starmap(autouri_rm, args)

    def get_lock(self, no_lock=False, timeout=None, poll_interval=None) -> BaseFileLock:
        """
        Args:
            no_lock: make it a dummy lock (for better code readability for context)
        """
        if no_lock:
            return contextmanager(lambda: (yield))()
        else:
            return self._get_lock(timeout=timeout, poll_interval=poll_interval)

    def localize_on(
        self,
        loc_prefix,
        recursive=False,
        make_md5_file=False,
        return_flag=False,
        depth=0,
        no_lock=False,
        no_checksum=False,
    ) -> Tuple[str, bool]:
        """Wrapper for classmethod localize().
        Localizes self on target directory loc_prefix.
        """
        return AutoURI.localize(
            src_uri=self,
            recursive=recursive,
            make_md5_file=make_md5_file,
            loc_prefix=loc_prefix,
            return_flag=return_flag,
            depth=depth,
            no_lock=no_lock,
            no_checksum=no_checksum,
        )

    @abstractmethod
    def _get_lock(self, timeout=None, poll_interval=None) -> BaseFileLock:
        """Locking mechanism with "with" context.

        Args:
            timeout: timeout in seconds
            poll_interval (float): polling interval in seconds
        """
        raise NotImplementedError

    @abstractmethod
    def get_metadata(self, skip_md5=False, make_md5_file=False) -> URIMetadata:
        """Metadata of a URI.
        This is more efficient than individually retrieving each item.
        md5 can be None. For example, HTTP URLs.

        Returns:
            exists:
            mtime: last modified time. seconds from the epoch
            size: bytes
            md5: md5 hexadecimal digest
        """
        raise NotImplementedError

    @abstractmethod
    def read(self, byte=False) -> Union[str, bytes]:
        """Reads string/byte from a URI."""
        raise NotImplementedError

    @abstractmethod
    def find_all_files(self, recursive=False) -> List[str]:
        """Recursively list all files (with full path/URI) on a directory.
        This method is available for a directory only and does not list sub-directories.
        """
        raise NotImplementedError

    @abstractmethod
    def _write(self, s):
        """Writes string/bytes to a URI. This is NOT protected by a locking mechanism.
        A file lock is already implemented in a higher level AutoURI.write().
        """
        raise NotImplementedError

    @abstractmethod
    def _rm(self):
        """Removes a URI from its storage. This is NOT protected by a locking mechanism.
        A file lock is already implemented in a higher level AutoURI.rm().
        """
        raise NotImplementedError

    @abstractmethod
    def _cp(self, dest_uri: Union[str, "AutoURI"]) -> bool:
        """Makes a copy on destination. This is NOT protected by a locking mechanism.
        Also, there is no checksum test for this function.
        A file lock/checksum is already implemented in a higher level AutoURI.cp().
        """
        raise NotImplementedError

    @abstractmethod
    def _cp_from(self, src_uri: Union[str, "AutoURI"]) -> bool:
        """Reversed version of "_cp".
        _cp is a binary operation so it can be defined in either source or destination
        URI class.

        A member function "_cp" is called first and if it returns False
        then this function will be called with reversed source and destination.

        This function is useful to be defined for user's custom classes inherited from
        AutoURI and existing URI classes.
        """
        raise NotImplementedError

    @classmethod
    def get_path_sep(cls) -> str:
        """Separator for directory."""
        return cls._PATH_SEP

    @classmethod
    def get_schemes(cls) -> Tuple[str, ...]:
        """Tuple of scheme strings.

        e.g. (gs://,), (s3://,), (http://, https://), tuple()
        """
        return cls._SCHEMES

    @classmethod
    def get_loc_suffix(cls) -> str:
        """File suffix for a MODIFIED file after recursive localization.
        This is required to distinguish a modified file from an original one.

        e.g. s3://temp1/tmp.json -> /scratch/cache_dir/tmp.s3.json
        """
        return cls._LOC_SUFFIX

    @classmethod
    def get_loc_prefix(cls) -> str:
        """Cache directory root path for localization.
        Tailing slash will be removed.
        """
        return cls.LOC_PREFIX.rstrip(cls.get_path_sep())

    @classmethod
    def localize(
        cls,
        src_uri,
        recursive=False,
        make_md5_file=False,
        loc_prefix=None,
        return_flag=False,
        depth=0,
        no_lock=False,
        no_checksum=False,
    ) -> Tuple[str, bool]:
        """Localize a source URI on this URI class (cls).

        Recursive localization is supported for the following file extensions:
            .json:
                Files defined only in values (not keys) can be recursively localized.
            .tsv/.csv:
                Files defined in all values can be recursively localized.
            Other extensions can be added through
                a global function init_uribase(loc_recurse_ext_and_fnc=).
                or a class constant LOC_RECURSE_EXT_AND_FNC directly.
                See loc_aux.py to define your own recursion function
                for a specific extension.
        Args:
            src_uri:
                Source URI
            recursive:
                Localize all files recursively in specified TEXT file extensions.
            make_md5_file:
                Make an md5 file on this if metadata doesn't have md5
                assuming that you have write permission on target's directory and
                its subdirectories recursively.
            loc_prefix:
                If defined, use it instead of cls.get_loc_prefix()
            return_flag:
                Returns a tuple of (localized uri string, flag)
                instead of localized uri string only
                    See "Returns" section for details about flag
            depth:
                To count recursion depth.
            no_lock:
                No file locking.
            no_checksum:
                Do not check md5 hash.
        Returns:
            loc_uri:
                Localized URI STRING (not a AutoURI instance) since it should be used
                for external function as a callback function.
            flag: (only if return_flag is on)
                Whether file is modified or localized on a different storage.
                "modified" means:
                    file contents are modified during recursive localization
                    so localized file is suffixed with
                    source's storage type. e.g. .s3, .gcs, and .local
                    not modified:
                "localized on a different storage" means:
                    file contents are NOT modified so localized file is not suffixed
                    and hence will keep the original file basename
                    but localiziation actually happened
                Otherwise, this flag will be False, which means that localization didn't
                happen because file's contents didn't change
                and it exists on the same storage so there is no need for localization.
                In this case, loc_uri will be identical to self._uri.
        """
        src_uri = AutoURI(src_uri)
        if not src_uri.is_valid:
            return (src_uri._uri, False) if return_flag else src_uri._uri

        if depth >= AutoURI.LOC_RECURSION_DEPTH_LIMIT:
            raise AutoURIRecursionError(
                "Maximum recursion depth {m} exceeded. "
                "Possible direct/indirect self-reference while "
                "recursive localization? related file: {f}".format(m=depth, f=src_uri)
            )

        if loc_prefix is None:
            loc_prefix = cls.get_loc_prefix()
        else:
            cls = AutoURI(loc_prefix).__class__
            loc_prefix = loc_prefix.rstrip(cls.get_path_sep())
        if not loc_prefix:
            raise ValueError("LOC_PREFIX is not defined.")

        # check if src and dest are on the same storage
        # to skip localization,
        # which means making a (possily modified) copy of original file
        on_different_storage = cls is not src_uri.__class__

        modified = False
        if recursive:

            def fnc_loc(uri):
                return cls.localize(
                    uri,
                    recursive=recursive,
                    make_md5_file=make_md5_file,
                    loc_prefix=loc_prefix,
                    return_flag=True,
                    depth=depth + 1,
                    no_lock=no_lock,
                    no_checksum=no_checksum,
                )

            for ext, fnc_recurse in AutoURI.LOC_RECURSE_EXT_AND_FNC.items():
                if src_uri.ext == ext:
                    # read source contents for recursive localization
                    src_contents = src_uri.read()
                    maybe_modified_contents, modified = fnc_recurse(
                        src_contents, fnc_loc
                    )
                    break

        if modified:
            # if modified, always suffix basename (before extension) with target storage cls
            basename = src_uri.basename_wo_ext + cls.get_loc_suffix() + src_uri.ext
            dirname = src_uri.loc_dirname
            loc_uri = cls.get_path_sep().join([loc_prefix, dirname, basename])
            AutoURI(loc_uri).write(maybe_modified_contents, no_lock=no_lock)

        elif on_different_storage:
            basename = src_uri.basename
            dirname = src_uri.loc_dirname

            loc_uri = cls.get_path_sep().join([loc_prefix, dirname, basename])
            src_uri.cp(
                dest_uri=loc_uri,
                make_md5_file=make_md5_file,
                no_lock=no_lock,
                no_checksum=no_checksum,
            )
        else:
            loc_uri = src_uri._uri

        return (loc_uri, modified or on_different_storage) if return_flag else loc_uri

    @staticmethod
    def init_uribase(
        md5_file_ext: Optional[str] = None,
        loc_recurse_ext_and_fnc: Optional[Dict[str, Callable]] = None,
        loc_recursion_depth_limit: Optional[int] = None,
        lock_file_ext: Optional[str] = None,
        lock_timeout: Optional[int] = None,
        lock_poll_interval: Optional[float] = None,
    ):
        if md5_file_ext is not None:
            URIBase.MD5_FILE_EXT = md5_file_ext
        if loc_recurse_ext_and_fnc is not None:
            URIBase.LOC_RECURSE_EXT_AND_FNC = loc_recurse_ext_and_fnc
        if loc_recursion_depth_limit is not None:
            URIBase.LOC_RECURSION_DEPTH_LIMIT = loc_recursion_depth_limit
        if lock_file_ext is not None:
            URIBase.LOCK_FILE_EXT = lock_file_ext
        if lock_timeout is not None:
            URIBase.LOCK_TIMEOUT = lock_timeout
        if lock_poll_interval is not None:
            URIBase.LOCK_POLL_INTERVAL = lock_poll_interval


class AutoURI(URIBase):
    """This class automatically detects and converts
    self into a URI class from a given URI string.
    It iterates over all IMPORTED URI subclasses and
    find the first one making it valid.

    This class can also work as an undefined URI class
    which can take/keep any type of variable and
    it's always read-only and not a valid URI.
    Therefore, you can use this class as a tester
    to check whether it's a valid URI or not.
    """

    def __init__(self, uri, thread_id=-1):
        super().__init__(uri, thread_id=thread_id)
        for c in URIBase.__subclasses__():
            if c is AutoURI:
                continue
            u = c(self._uri, thread_id=thread_id)
            if u.is_valid:
                self.__class__ = c
                self._uri = u._uri
                return

    def _get_lock(self, timeout=None, poll_interval=None):
        self.__raise_value_error()

    def get_metadata(self, skip_md5=False, make_md5_file=False):
        self.__raise_value_error()

    def read(self, byte=False):
        self.__raise_value_error()

    def find_all_files(self):
        self.__raise_value_error()

    def rmdir(self):
        self.__raise_value_error()

    def _write(self, s):
        self.__raise_value_error()

    def _rm(self):
        self.__raise_value_error()

    def _cp(self, dest_uri):
        self.__raise_value_error()

    def _cp_from(self, src_uri):
        self.__raise_value_error()

    def __raise_value_error(self):
        raise ValueError("Not a valid URI?. {f}".format(f=self._uri))



import os
import time
from typing import Any, Tuple

import pytest

from autouri.abspath import AbsPath, convert_relpath_to_abspath_if_valid
from autouri.autouri import AutoURI, URIBase
from autouri.httpurl import ReadOnlyStorageError

from .files import (
    common_paths,
    make_files_in_dir,
    recurse_raise_if_uri_not_exist,
    v6_txt_contents,
)


@pytest.mark.parametrize(
    "path,converted",
    [
        ("/asdfasf/sdaf/", "/asdfasf/sdaf/"),
        ("LICENSE", "LICENSE"),
        ("setup.py", "setup.py"),
        ("setup.py.", "setup.py."),
        ("license", "license"),
        ("does-not-exist.json", "does-not-exist.json"),
        ("~/sadfsf", "~/sadfsf"),
        ("/etc/passwd", "/etc/passwd"),
        ("conftest.py", os.getcwd() + "/conftest.py"),
        ("files.py", os.getcwd() + "/files.py"),
    ],
)
def test_convert_relpath_to_abspath_if_valid(path, converted) -> Any:
    assert convert_relpath_to_abspath_if_valid(path, allowed_exts=(".py")) == converted


@pytest.mark.parametrize("path", common_paths())
def test_abspath_uri(path) -> Any:
    assert AbsPath(path).uri == os.path.expanduser(path)


@pytest.mark.parametrize("path", common_paths())
def test_abspath_uri_wo_ext(path) -> str:
    assert AbsPath(path).uri_wo_ext == os.path.splitext(os.path.expanduser(path))[0]


@pytest.mark.parametrize("path", common_paths())
def test_abspath_uri_wo_scheme(path) -> str:
    assert AbsPath(path).uri_wo_scheme == os.path.expanduser(path)


@pytest.mark.parametrize("path", common_paths())
def test_abspath_is_valid(path) -> bool:
    """Also tests AutoURI auto-conversion since it's based on is_valid property"""
    expected = os.path.isabs(os.path.expanduser(path))
    assert AbsPath(path).is_valid == expected
    assert not expected or type(AutoURI(path)) == AbsPath


@pytest.mark.parametrize("path", common_paths())
def test_abspath_dirname(path) -> str:
    assert AbsPath(path).dirname == os.path.dirname(os.path.expanduser(path))


@pytest.mark.parametrize("path", common_paths())
def test_abspath_dirname_wo_scheme(path) -> str:
    assert AbsPath(path).dirname_wo_scheme == os.path.dirname(os.path.expanduser(path))


@pytest.mark.parametrize("path", common_paths())
def test_abspath_loc_dirname(path) -> str:
    assert AbsPath(path).loc_dirname == os.path.dirname(os.path.expanduser(path)).strip(
        os.path.sep
    )


@pytest.mark.parametrize("path", common_paths())
def test_abspath_basename(path) -> str:
    assert AbsPath(path).basename == os.path.basename(os.path.expanduser(path))


@pytest.mark.parametrize("path", common_paths())
def test_abspath_basename_wo_ext(path) -> str:
    assert (
        AbsPath(path).basename_wo_ext
        == os.path.splitext(os.path.basename(os.path.expanduser(path)))[0]
    )


@pytest.mark.parametrize("path", common_paths())
def test_abspath_ext(path) -> str:
    assert AbsPath(path).ext == os.path.splitext(os.path.expanduser(path))[1]


def test_abspath_exists(local_v6_txt):
    assert AbsPath(local_v6_txt).exists
    assert not AbsPath(local_v6_txt + ".should-not-be-here").exists
    assert not AbsPath("/hey/this/should/not/be/here.txt").exists


def test_abspath_mtime(local_v6_txt):
    u = AbsPath(local_v6_txt + ".tmp")
    u.write("temp file for testing")
    now = time.time()
    assert now - 10 < u.mtime < now + 10
    u.rm()
    assert not u.exists


def test_abspath_size(local_v6_txt, v6_txt_size):
    assert AbsPath(local_v6_txt).size == v6_txt_size


def test_abspath_md5(local_v6_txt, v6_txt_md5_hash):
    assert AbsPath(local_v6_txt).md5 == v6_txt_md5_hash


def test_abspath_md5_from_file(local_v6_txt, v6_txt_md5_hash):
    u_md5 = AbsPath(local_v6_txt + URIBase.MD5_FILE_EXT)
    if u_md5.exists:
        u_md5.rm()
    assert not u_md5.exists
    u = AbsPath(local_v6_txt)
    assert u.md5_from_file is None

    u.get_metadata(make_md5_file=True)
    assert u_md5.exists
    assert u.md5_from_file == v6_txt_md5_hash
    u_md5.rm()
    assert not u_md5.exists


def test_abspath_md5_file_uri(local_v6_txt):
    assert (
        AbsPath(local_v6_txt + URIBase.MD5_FILE_EXT).uri
        == local_v6_txt + URIBase.MD5_FILE_EXT
    )


def test_abspath_cp_url(local_v6_txt, url_test_path) -> "AutoURI":
    """Test copying local_v6_txt to the following destination storages:
    url_test_path: local -> url
        This will fail as intended since URL is read-only.
    """
    u = AbsPath(local_v6_txt)
    basename = os.path.basename(local_v6_txt)

    for test_path in (url_test_path,):
        u_dest = AutoURI(os.path.join(test_path, "test_abspath_cp", basename))
        with pytest.raises(ReadOnlyStorageError):
            _, ret = u.cp(u_dest, return_flag=True)


def test_abspath_cp(
    local_v6_txt, local_test_path, s3_test_path, gcs_test_path, url_test_path
) -> "AutoURI":
    """Test copying local_v6_txt to the following destination storages:
        local_test_path: local -> local
        s3_test_path: local -> s3
        gcs_test_path: local -> gcs

    Parameters to be tested:
        no_lock:
            Copy with no locking mechanism. There is no way to test this thoroughly here.
            This will be tested with multiple threads later in test_rece_cond.py.
        no_checksum:
            Don't check md5-hash/size/mtime to skip copying (even if file already exists on destination).
        make_md5_file:
            Make md5 file on destination only when it's REQUIRED.
            It's required only if we need to compare md5 hash of source and target.
    """
    u = AbsPath(local_v6_txt)
    basename = os.path.basename(local_v6_txt)

    for test_path in (local_test_path, s3_test_path, gcs_test_path):
        u_dest = AutoURI(os.path.join(test_path, "test_abspath_cp", basename))
        if u_dest.exists:
            u_dest.rm()

        assert not u_dest.exists
        _, ret = u.cp(u_dest, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 0
        u_dest.rm()

        assert not u_dest.exists
        # cp without lock will be tested throughly in test_race_cond.py
        _, ret = u.cp(u_dest, no_lock=True, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 0
        u_dest.rm()

        # trivial: copy without checksum when target doesn't exists
        assert not u_dest.exists
        _, ret = u.cp(u_dest, no_checksum=True, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 0

        # copy without checksum when target exists
        m_dest = u_dest.get_metadata()
        assert m_dest.exists
        time.sleep(1)
        _, ret = u.cp(u_dest, no_checksum=True, return_flag=True)
        # compare new mtime vs old mtime
        # new time should be larger if it's overwritten as intended
        assert u_dest.mtime > m_dest.mtime and u.read() == u_dest.read() and ret == 0

        # copy with checksum when target exists
        m_dest = u_dest.get_metadata()
        assert m_dest.exists
        _, ret = u.cp(u_dest, return_flag=True)
        # compare new mtime vs old mtime
        # new time should be the same as old time
        assert u_dest.mtime == m_dest.mtime and u.read() == u_dest.read() and ret == 1

        # make_md5_file works only when it's required
        # i.e. when we need to compare md5 hash of src vs target
        # so target must exist prior to test it
        assert u_dest.exists
        # delete md5 file if exists
        u_dest_md5_file = AutoURI(u_dest.uri + URIBase.MD5_FILE_EXT)
        if u_dest_md5_file.exists:
            u_dest_md5_file.rm()
        _, ret = u.cp(u_dest, make_md5_file=True, return_flag=True)
        assert u_dest.exists and u.read() == u_dest.read() and ret == 1
        u_dest.rm()


def test_abspath_write(local_test_path):
    u = AbsPath(local_test_path + "/test_abspath_write.tmp")

    assert not u.exists
    u.write("test")
    assert u.exists and u.read() == "test"
    u.rm()

    # this will be tested more with multiple threads in test_race_cond.py
    assert not u.exists
    u.write("test2", no_lock=True)
    assert u.exists and u.read() == "test2"
    u.rm()
    assert not u.exists


def test_abspath_write_no_permission():
    u = AbsPath("/test-permission-denied/x.tmp")
    with pytest.raises(PermissionError):
        u.write("test")


def test_abspath_rm(local_test_path):
    u = AbsPath(local_test_path + "/test_abspath_rm.tmp")

    assert not u.exists
    u.write("")
    assert u.exists
    u.rm()
    assert not u.exists

    # this will be tested more with multiple threads in test_race_cond.py
    assert not u.exists
    u.write("", no_lock=True)
    assert u.exists
    u.rm()
    assert not u.exists


def test_abspath_get_metadata(local_v6_txt, v6_txt_size, v6_txt_md5_hash):
    u = AbsPath(local_v6_txt)

    m1 = u.get_metadata()
    assert m1.md5 == v6_txt_md5_hash
    assert m1.size == v6_txt_size

    m2 = u.get_metadata(skip_md5=True)
    assert m2.md5 is None
    assert m2.size == v6_txt_size

    u_md5 = AbsPath(local_v6_txt + ".md5")
    if u_md5.exists:
        u_md5.rm()
    m3 = u.get_metadata(make_md5_file=True)
    assert m3.md5 == v6_txt_md5_hash
    assert m3.size == v6_txt_size
    assert u_md5.exists
    assert u_md5.read() == v6_txt_md5_hash


def test_abspath_read(local_v6_txt):
    u = AbsPath(local_v6_txt)
    assert u.read() == v6_txt_contents()
    assert u.read(byte=True) == v6_txt_contents().encode()


def test_abspath_find_all_files(local_test_path):
    """Make a directory structure with empty files and an empty directory.

    Check if find_all_files() returns correct file (not sub-directory) paths.
    Also check if find_all_files() does not return an empty sub-directory.
    """
    prefix = os.path.join(local_test_path, "test_abspath_find_all_files")
    all_files = make_files_in_dir(prefix, make_local_empty_dir_d_a=True)
    empty_sub_dir = os.path.join(prefix, "d", "a")
    assert os.path.exists(empty_sub_dir) and os.path.isdir(empty_sub_dir)

    all_files_found = AbsPath(prefix).find_all_files()
    assert sorted(all_files_found) == sorted(all_files)
    for file in all_files:
        assert AbsPath(file).exists
        assert file.rstrip("/") != empty_sub_dir.rstrip("/")


def test_abspath_rmdir(local_test_path):
    """Make a directory structure with empty files and an empty directory.

    Check if rmdir() deletes the root directory itself including
    all empty files and empty directory on given $prefix.
    """
    prefix = os.path.join(local_test_path, "test_abspath_rmdir")
    all_files = make_files_in_dir(prefix, make_local_empty_dir_d_a=True)
    empty_sub_dir = os.path.join(prefix, "d", "a")
    assert os.path.exists(empty_sub_dir) and os.path.isdir(empty_sub_dir)

    # test rmdir(dry_run=True)
    AbsPath(prefix).rmdir(dry_run=True)
    for file in all_files:
        assert AbsPath(file).exists

    # test rmdir(dry_run=False)
    AbsPath(prefix).rmdir(dry_run=False)
    for file in all_files:
        assert not AbsPath(file).exists
    assert not os.path.exists(empty_sub_dir)
    assert not os.path.exists(prefix)


# original methods in AbsPath
def test_abspath_get_mapped_url(local_v6_txt):
    u = AbsPath(local_v6_txt)
    dirname = os.path.dirname(local_v6_txt)
    basename = os.path.basename(local_v6_txt)
    url_prefix = "http://my.test.com"

    AbsPath.init_abspath(map_path_to_url={dirname: url_prefix})
    assert u.get_mapped_url() == os.path.join(url_prefix, basename)

    AbsPath.init_abspath(map_path_to_url=dict())
    assert u.get_mapped_url() is None


def test_abspath_mkdirname(local_test_path):
    f = os.path.join(local_test_path, "test_abspath_mkdirname", "tmp.txt")
    AbsPath(f).mkdir_dirname()
    assert os.path.exists(os.path.dirname(f))


def test_abspath_soft_link(local_test_path, local_v6_txt):
    u_src = AbsPath(local_v6_txt)
    f = os.path.join(local_test_path, "test_abspath_soft_link", "v6.txt")
    u_target = AbsPath(f)
    u_target.mkdir_dirname()
    u_src.soft_link(u_target)
    assert u_target.exists and u_target.read() == v6_txt_contents()
    assert u_src.uri == os.path.realpath(u_target.uri)

    with pytest.raises(OSError):
        # file already exists
        u_src.soft_link(u_target)
    # no error if force
    u_src.soft_link(u_target, force=True)
    u_target.rm()


# staticmethods
def test_abspath_get_abspath_if_exists():
    # write a local file on CWD.
    test_local_file_abspath = os.path.join(os.getcwd(), "test.txt")
    u = AbsPath(test_local_file_abspath)
    if u.exists:
        u.rm()

    # if it doesn't exist
    assert AbsPath.get_abspath_if_exists("test.txt") == "test.txt"
    assert AbsPath.get_abspath_if_exists(AutoURI("test.txt")) == "test.txt"

    u.write("hello-world")

    # if it exists
    assert AbsPath.get_abspath_if_exists("test.txt") == test_local_file_abspath
    assert AbsPath.get_abspath_if_exists(AutoURI("test.txt")) == test_local_file_abspath

    assert AbsPath.get_abspath_if_exists("tttttttttest.txt") == "tttttttttest.txt"
    assert (
        AbsPath.get_abspath_if_exists(AutoURI("tttttttttest.txt")) == "tttttttttest.txt"
    )
    assert (
        AbsPath.get_abspath_if_exists("~/if-it-does-not-exist")
        == "~/if-it-does-not-exist"
    )
    assert AbsPath.get_abspath_if_exists("non-existing-file") == "non-existing-file"

    u.rm()


# classmethods
def test_abspath_get_path_sep() -> str:
    assert AbsPath.get_path_sep() == os.path.sep


def test_abspath_get_schemes() -> Tuple[str, ...]:
    assert AbsPath.get_schemes() == tuple()


def test_abspath_get_loc_suffix() -> str:
    assert AbsPath.get_loc_suffix() == ".local"


def test_abspath_get_loc_prefix() -> str:
    test_loc_prefix = "test_abspath_get_loc_prefix"
    AbsPath.init_abspath(loc_prefix=test_loc_prefix)
    assert AbsPath.get_loc_prefix() == test_loc_prefix
    AbsPath.init_abspath(loc_prefix="")
    assert AbsPath.get_loc_prefix() == ""


def test_abspath_localize(
    local_test_path,
    local_j1_json,
    local_v41_json,
    local_v421_tsv,
    local_v5_csv,
    local_v6_txt,
    s3_j1_json,
    s3_v41_json,
    s3_v421_tsv,
    s3_v5_csv,
    s3_v6_txt,
    gcs_j1_json,
    gcs_v41_json,
    gcs_v421_tsv,
    gcs_v5_csv,
    gcs_v6_txt,
    url_j1_json,
    url_v41_json,
    url_v421_tsv,
    url_v5_csv,
    url_v6_txt,
) -> Tuple[str, bool]:
    """Recursive localization is supported for the following file extensions:
        .json:
            Files defined only in values (not keys) can be recursively localized.
        .tsv/.csv:
            Files defined in all values can be recursively localized.

    This function will test localizing j1.json file on each remote storage.
    This JSON file has file paths including .tsv and .csv, which also include
    other files in its contents.
    Therefore, when the recursive flag is on, all files in these JSON, TSV, CSV
    files should be localized recursively with correct file names
    (controlled by cls.loc_prefix and cls.loc_suffix).

    Filenaming for (recursive) localization:
        cls.loc_prefix + remote_file_path_without_scheme + cls.loc_suffix (for recursvely only)

    For example,
    s3://test-bucket/j1.json has some file paths on s3://.

    With recursive localization, all these files must be localized on /tmp/user/loc_prefix/ with
    a correct directory structure (keeping original structure on source: i.e. bucket name, path)
    and the name of the JSON file should be j1.local.json since contents of this file should be
    modified to point to localized files in it. This is recursively done for all files in it too.

    Without recursive localization, autouri doesn't look inside that JSON file and just localize
    the file itself alone on /tmp/user/loc_prefix/ while keeping the same filename j1.local.json.

    Test localizing on a local storage from the following remote storages:
        local_test_path: local -> local
        s3_test_path: s3 -> local
        gcs_test_path: gcs -> local
        url_test_path: url -> local

    Parameters to be tested:
        make_md5_file:
            Make md5 file on destination only when it's REQUIRED.
            It's required only if we need to compare md5 hash of source and target.
            This is already tested cp and it's actually needed for local storage.
            Cloud URIs will provide md5 hash info in their metadata so md5 file
            is not required and hence will not be created even with this flag on.
        recursive:
            j1.json
    """
    loc_prefix = os.path.join(local_test_path, "test_abspath_localize")

    for j1_json in (local_j1_json,):
        # localization from local storage
        u_j1_json = AutoURI(j1_json)
        loc_prefix_ = loc_prefix + u_j1_json.__class__.get_loc_suffix()

        # for localization both with or without recursive
        # nothing should be localized actually
        # since they are already on a local storage
        # so loc_prefix directory itself shouldn't be created
        loc_uri, localized = AbsPath.localize(
            u_j1_json, recursive=False, return_flag=True, loc_prefix=loc_prefix_
        )
        assert loc_uri == u_j1_json.uri and not localized
        assert not os.path.exists(loc_prefix)

        loc_uri, localized = AbsPath.localize(
            u_j1_json, recursive=True, return_flag=True, loc_prefix=loc_prefix_
        )
        assert loc_uri == u_j1_json.uri and not localized
        assert not os.path.exists(loc_prefix)
        # check if all URIs defeind in localized JSON file exist
        recurse_raise_if_uri_not_exist(loc_uri)

    # localization from remote storages
    for j1_json in (gcs_j1_json, s3_j1_json, url_j1_json):
        u_j1_json = AutoURI(j1_json)
        loc_prefix_ = loc_prefix + u_j1_json.__class__.get_loc_suffix()

        loc_uri, localized = AbsPath.localize(
            u_j1_json, recursive=False, return_flag=True, loc_prefix=loc_prefix_
        )
        assert loc_uri == os.path.join(
            loc_prefix_, u_j1_json.loc_dirname, u_j1_json.basename
        )
        assert localized and os.path.exists(loc_uri)

        loc_uri, localized = AbsPath.localize(
            u_j1_json, recursive=True, return_flag=True, loc_prefix=loc_prefix_
        )
        assert loc_uri == os.path.join(
            loc_prefix_,
            u_j1_json.loc_dirname,
            u_j1_json.basename_wo_ext + AbsPath.get_loc_suffix() + u_j1_json.ext,
        )
        assert localized and os.path.exists(loc_uri)
        # check if all URIs defeind in localized JSON file exist
        recurse_raise_if_uri_not_exist(loc_uri)



from .abspath import AbsPath
from .autouri import AutoURI, URIBase
from .gcsuri import GCSURI
from .httpurl import HTTPURL
from .s3uri import S3URI

__all__ = ["AbsPath", "AutoURI", "URIBase", "GCSURI", "HTTPURL", "S3URI"]
__version__ = "0.4.2"
